# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import importlib

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from hsr1.utils.spectrum import spectrumUtils as SpectrumUtils
from hsr1.utils.spectrum.pixelSpectrum import PixelSpectrum
from hsr1.utils.spectrum.weightedSpectrum import WeightedSpectrum


class RawDataset:
    def __init__(self, dataset:pd.DataFrame, 
                 calibrations:dict=None, calibrations_filepath=None, 
                 amplitude_calibration=None, amplitude_calibration_filepath=None, 
                 spectrum_type="wavelength"):
        """stores a raw dataset containing all the channels of raw data
        
        params:
            dataset: dataframe with columns "pc_time_end_measurement", "channel_0", "channel_1" etc.
            calibrations: dict containing polynomials with keys "channel_0", "channel_1" etc.
            calibrations_filepath: filepath to the current wavelength calibration file
            amplitude_cal: an amplitude calibration to apply to found spectrum, only used for graphs
            amplitude_calibration_filepath
        """
        self.dataset = dataset
        self.calibrations = calibrations
        self.amplitude_calibration = amplitude_calibration
        self.spectrum_type = spectrum_type
        
        if spectrum_type != "wavelength" and spectrum_type != "pixel":
            raise ValueError("spectrum_type must be either \"wavelength\" or \"pixel\"")
        
        if calibrations is None and spectrum_type == "wavelength":
            if calibrations_filepath is None:
                raise ValueError("""original calibrations must be passed, either as a dict or as the location of the calibration file. If this is uncalibrated data, pass spectrum_type=\"pixel\"""")
            self.calibrations = SpectrumUtils.read_wavelength_polys(calibrations_filepath)
        
        if spectrum_type == "pixel":
            if calibrations is not None or calibrations_filepath is not None:
                print("there is no point passing existing calibrations to a pixel dataset, pixel readings are assumed to be uncalibrated. \
                      \n if this is a calibration that is used on this spectrometer, you can change the default_calibration parameter in find_all_calibrations.")
        
        if amplitude_calibration is None and amplitude_calibration_filepath is not None:
            self.amplitude_calibration = SpectrumUtils.read_amplitude_file(amplitude_calibration_filepath)
        
        if amplitude_calibration is None and amplitude_calibration_filepath is None:
            self.amplitude_calibration = np.ones(801)
        
    
    
    def plot_calibration(self, polys, names=None):
        """plots a line graph of pixel against wavelength for each channel
        params:
            polys: list of lists or dict containing the polynommials for each channel's new calibration
            names: list of names to display on the legend
        """
        ps = PixelSpectrum()
        
        
        if isinstance(polys, dict):
            poly_names = []
            poly_list = []
            for poly in polys:
                poly_list.append(polys[poly])
                poly_names.append(poly)
            polys = poly_list 
            names = poly_names
        
        if names is None:
            names = [""]*len(polys)
        
        plt.figure(figsize=(20,10))
        for poly, name in zip(polys, names):
            plt.plot(ps.apply_wavelength_calibration(np.arange(0, 1200), poly, inverse=True), np.arange(300, 1101), linewidth=0.5, label=name + ": "+SpectrumUtils.poly_to_string(poly))
        plt.xlim(0, 1200)
        plt.suptitle("individual channel calibrations")
        plt.xlabel("pixels")
        plt.ylabel("wavelength")
        plt.legend()
        plt.show(block=False)
    
    def plot_calibration_wide(self, polys, names=None, title=""):
        """plots a line graph of pixel against wavelength for each channel
        params:
            polys: list of lists containing the polynommials for each channel's new calibration
            names: list of names to display on the legend
        """
        if names is None:
            names = [""]*len(polys)
        
        plt.figure(figsize=(20,10))
        for poly, name in zip(polys, names):
            x = np.linspace(-10000, 10000, 1000)
            y = [SpectrumUtils.apply_polynomial(i, poly[0], poly[1], poly[2], poly[3]) for i in x]
            plt.plot(x, y, label=name + ": " + SpectrumUtils.poly_to_string(poly))

        plt.axvline(0)
        plt.axvline(1200)
        plt.suptitle(title)
        plt.legend()
        plt.show(block=False)
    
    def plot_all_errors(self, polys, dips, names):
        plt.figure(figsize=(20, 10))
        for i, poly, _dips, name in zip(list(range(len(polys))), polys, dips, names):
            measured = _dips[0]
            ref = _dips[1]
            error = measured-ref
            
            plt.plot(ref, error-(i/50), label=name)
        
        plt.legend()
        plt.show(block=False)
    
    
    def find_all_calibrations(self, file_output=None, 
                              reference_spectrum=None, reference_filepath=None, 
                              plot_total_output=True, plot_individual_results=False, plot_debug=False, 
                              **kwargs):
        """loops through each channel in the dataset and finds the correct calibration
        params:
            file_output: for outputting results, which text file to use. leave as None to not output
            reference_spectrum: a spectrum in wavelength space that the measured data will be matched to
            plot_total_output: plots the calibration for each channel on one graph
            plot_individual_results: plots the results from each channel, passes plot_results to each calibration
            plot_debug: plots the result from each segment of each channel, passed to each calibration
            kwargs: arguments to pass to each channel's find_calibration method, cleaner than having all arguments as parameters here
        """
        if reference_filepath is None:
            res = importlib.resources.files("hsr1.data").joinpath("smarts.txt")
            file = importlib.resources.as_file(res)
            with file as f:
                reference_filepath = f
            
        
        
        if reference_spectrum is None:
            try:
                reference_spectrum = SpectrumUtils.read_simple_file(reference_filepath)
            except ValueError:
                reference_spectrum = SpectrumUtils.read_reference_file(reference_filepath)
        
        
        all_dips = []
        polys = []
        names = []
        new_calibrations = {}
        for channel_name in self.dataset.columns[1:]:
            
            
            ##### extract the polynomial that has already been applied to this channel
            applied_poly = None
            if self.spectrum_type == "wavelength":
                if channel_name in self.calibrations:
                    applied_poly = self.calibrations[channel_name]
                else:
                    print("error while calibrating "+channel_name+": couldnt find an existing calibration for this channel")
                    continue
            
            ps = PixelSpectrum(cal_poly=applied_poly)
            
            channel = self.dataset[["pc_time_end_measurement", channel_name]].copy()
            channel = channel.rename(columns={channel_name: "global_spectrum"})
            
            
            
            ##### calculate one pixel spectrum, for graphing
            pixel_spectrum = None
            if self.spectrum_type == "wavelength":
                pixel_data = ps.calculate_from_wavelengths(channel, ["global_spectrum"])
                max_index = 0
                if len(channel) > 1:
                    integrals = pixel_data["global_spectrum"].apply(np.sum)
                    max_index = np.argmax(integrals)
                pixel_spectrum = pixel_data.loc[max_index, "global_spectrum"]
            elif self.spectrum_type == "pixel":
                max_index = 0
                if len(channel) > 1:
                    integrals = channel.apply(np.sum)
                    max_index = np.argmax(integrals)
                pixel_spectrum = channel.loc[max_index, "global_spectrum"]
            
            if max(pixel_spectrum) == 0:
                print("invalid spectrum selected for graphing")
            
            ##### calculate the weighted spectrum and calibrate
            ws = None
            if self.spectrum_type == "wavelength":
                ws = WeightedSpectrum(wavelength_spectra=channel, reference_spectrum=reference_spectrum, 
                                      applied_poly=applied_poly, 
                                      filter_low_values=False, 
                                      amplitude_calibration=self.amplitude_calibration, 
                                      pixel_spectrum=pixel_spectrum, 
                                      name=channel_name, **kwargs)
            elif self.spectrum_type == "pixel":
                ws = WeightedSpectrum(pixel_spectra=channel, reference_spectrum=reference_spectrum, 
                                      applied_poly=applied_poly, 
                                      filter_low_values=False, 
                                      amplitude_calibration=self.amplitude_calibration, 
                                      pixel_spectrum=pixel_spectrum, 
                                      name=channel_name, **kwargs)
                
            wavelength_cal, dips = (None, None)
            if plot_total_output:
                wavelength_cal, dips = ws.find_calibration(plot_debug=plot_debug, plot_results=plot_individual_results, 
                                                            return_dips=plot_total_output, 
                                                            **kwargs)
            else:
                wavelength_cal = ws.find_calibration(plot_debug=plot_debug, plot_results=plot_individual_results, 
                                                            return_dips=plot_total_output, 
                                                            **kwargs)
            if wavelength_cal is not None:
                all_dips.append(dips)
                new_calibrations[channel_name] = wavelength_cal
                polys.append(wavelength_cal)
                names.append(channel_name)
        
        if plot_total_output:
            self.plot_calibration(polys, names)
            self.plot_calibration_wide(polys, names)
            self.plot_all_errors(polys, all_dips, names)
        
        if file_output is not None:
            SpectrumUtils.format_output(new_calibrations, file_output)
        
        return new_calibrations


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import uuid
import os
import importlib

import numpy as np
import pandas as pd

pvlib_imported = False

try:
    import pvlib
    pvlib_imported = True
except:
    print("pvlib was not successfully imported")

import hsr1
from hsr1.utils.spectrum import spectrumUtils

SCALING_FACTOR = 0.617

rand = np.random.default_rng()

class SyntheticDataset():
    def __init__(self):
        if not pvlib_imported:
            raise ImportError("pvlib is required to generate synthetic datasets")
    
    def __init__(self, spectral_func=None, 
                 start_date="2024-01-01 00:00:00+00:00", 
                 end_date="2024-01-02 00:00:00+00:00",
                 latitude=53.14,
                 longitude=-1.64,
                 altitude=40):
        
        
        if spectral_func is None:
            self.spectral_func = self.generate_spectral_data
        
        self.spectral_data = None
        self.system_data = None
        self.deployment_metadata = None
        self.accessory_data = None
        
        self.start_date = start_date
        self.end_date = end_date
        self.latitude = latitude
        self.longitude = longitude
        self.altitude = altitude
        
        self.dataseries_id = None
    
    def get_dfs(self):
        return [self.spectral_data, self.system_data, self.deployment_metadata, self.accessory_data]
    
    def generate_custom_column(self, table, column, function, *args, **kwargs):
        if table == "spectral_data":
            self.spectral_data[column] = function(self.spectral_data["pc_time_end_measurement"], 
                                                  *args, **kwargs)
        elif table == "system_data":
            self.system_data[column] = function(self.spectral_data["pc_time_end_measurement"], 
                                                *args, **kwargs)
        elif table == "accessory_data":
            self.accessory_data[column] = function(self.accessory_data["pc_time_end_measurement"], 
                                                   *args, **kwargs)
        
        elif table == "deployment_metadata":
            self.deployment_metadata[column] = function(*args, **kwargs)
            
        else:
            raise ValueError("table not found")
    
    def remove_rows(self, table, function, *args, **kwargs):
        if table == "spectral_data":
            self.spectral_data = self.spectral_data.loc[function(self.spectral_data["pc_time_end_measurement"], *args, **kwargs), :]
        if table == "system_data":
            self.system_data = self.system_data.loc[function(self.system_data["pc_time_end_measurement"], *args, **kwargs), :]
        if table == "accessory_data":
            self.accessory_data = self.accessory_data.loc[function(self.accessory_data["pc_time_end_measurement"], *args, **kwargs), :]

    def global_integral(self, timestamps):
        if self.system_data is None:
            raise ValueError("system_data is required, generate that first")
            
        
        
    
    def generate_spectral_data(self, period="min"):
        spectral_data = pd.DataFrame()
        
        spectral_data["pc_time_end_measurement"] = pd.date_range(self.start_date, self.end_date, freq=period).astype(str)
        num_readings = len(spectral_data)
        spectral_data["sample_id"] = self.generate_sample_ids(spectral_data["pc_time_end_measurement"])
        spectral_data["dataseries_id"] = self.generate_dataseries_id()
        spectral_data["global_spectrum"] = self.smarts_integral_data(spectral_data["pc_time_end_measurement"]).tobytes()
        spectral_data["diffuse_spectrum"] = self.smarts_integral_data(spectral_data["pc_time_end_measurement"]).tobytes()
        spectral_data["global_integral"] = pd.to_datetime(spectral_data["pc_time_end_measurement"]).dt.hour
        spectral_data["diffuse_integral"] = spectral_data["global_integral"]/2
        spectral_data["global_molar"] = spectral_data["global_integral"]
        spectral_data["diffuse_molar"] = spectral_data["diffuse_integral"]
        spectral_data["camera_temp"] = rand.normal(30, 0.1, num_readings)
        
        
        self.spectral_data = spectral_data
    
    def generate_system_data(self, period="min"):
        system_data = pd.DataFrame()
        
        system_data["pc_time_end_measurement"] = pd.date_range(self.start_date, self.end_date, freq=period).astype(str)
        system_data["gps_time"] = pd.to_datetime(system_data["pc_time_end_measurement"]).dt.tz_localize(None).astype(str)
        system_data["gps_status"] = "A"
        system_data["gps_latitude"] = self.latitude
        system_data["gps_longitude"] = self.longitude
        system_data["gps_altitude"] = 100
        system_data["pressure"] = 1000
        system_data["baro_temp"] = 25
        system_data["rh"] = 13
        system_data["rh_temp"] = 0.0
        system_data["dataseries_id"] = self.generate_dataseries_id()
        system_data["sample_id"] = self.generate_sample_ids(system_data["pc_time_end_measurement"])
        
        
        self.system_data = system_data
    
    
    def generate_system_data_accessory(self, accessory_data=None, spectral_timestamps=None, spectral_sample_ids=None):
        if accessory_data is None:
            accessory_data = self.accessory_data
        if spectral_timestamps is None:
            spectral_timestamps = self.spectral_data["pc_time_end_measurement"]
        if spectral_sample_ids is None:
            spectral_sample_ids = self.spectral_data["sample_id"]
        
        system_data = hsr1.utils.ReformatData().reformat_system_data_reformatted_accessory(accessory_data, spectral_timestamps, spectral_sample_ids)
        
        system_data["dataseries_id"] = self.generate_dataseries_id()
        system_data["sample_id"] = self.generate_sample_ids(system_data["pc_time_end_measurement"])
        
        self.system_data = system_data
        
        
    
    def generate_deployment_metadata(self, filepath):
        self.deployment_metadata = hsr1.utils.ReformatData().reformat_deployment_metadata(filepath)
        # self.deployment_metadata["default_elevation"] = str(self.altitude)
        # self.deployment_metadata["default_latitude"] = str(self.latitude)
        # self.deployment_metadata["default_longitude"] = str(self.longitude)
    
    def generate_accessory_data(self, period="min"):
        accessory_data = pd.DataFrame()
        
        accessory_data["pc_time_end_measurement"] = pd.date_range(self.start_date, self.end_date, freq=period).astype(str)
        
        num_readings = len(accessory_data)
        
        accessory_data["Clock_Error"] = 0.0
        accessory_data["Latitude"] = rand.normal(self.latitude, 0.1, num_readings)
        accessory_data["Longitude"] = rand.normal(self.longitude, 0.1, num_readings)
        accessory_data["Altitude"] = rand.normal(100.0, 0.1, num_readings)
        accessory_data["GPSAge"] = 0.0
        accessory_data["NSV"] = rand.normal(10.0, 0.1, num_readings).astype(int)
        accessory_data["GPSSpare"] = rand.normal(5.0, 0.1, num_readings)
        accessory_data["_15Vin"] = rand.normal(23000.0, 0.1, num_readings)
        accessory_data["_3VCAP"] = rand.normal(22000.0, 0.1, num_readings)
        accessory_data["HTRIn"] = rand.normal(50.0, 0.1, num_readings)
        accessory_data["Vcc"] = rand.normal(3300.0, 0.1, num_readings)
        accessory_data["VSpare"] = rand.normal(0.0, 0.1, num_readings)
        accessory_data["I_Tot"] = rand.normal(484.0, 0.1, num_readings)
        accessory_data["I_15VPC"] = rand.normal(200.0, 0.1, num_readings)
        accessory_data["I_15VCAM"] = rand.normal(100.0, 0.1, num_readings)
        accessory_data["I_5VPC"] = rand.normal(4.0, 0.1, num_readings)
        accessory_data["ISpare"] = 0.0
        accessory_data["T_CPU"] = rand.normal(30.0, 0.1, num_readings)
        accessory_data["T_Bezel"] = rand.normal(29.0, 0.1, num_readings)
        accessory_data["T_RH"] = 0.0
        accessory_data["T_Baro"] = rand.normal(27.0, 0.1, num_readings)
        accessory_data["RH"] = rand.normal(12.0, 0.1, num_readings)
        accessory_data["Pressure"] = rand.normal(900.0, 0.1, num_readings)
        accessory_data["IMU_Output_Type"] = 2.0
        accessory_data["Yaw"] = rand.normal(358.0, 0.1, num_readings)
        accessory_data["Roll"] = rand.normal(-1.62, 0.1, num_readings)
        accessory_data["Pitch"] = rand.normal(10.31, 0.1, num_readings)
        accessory_data["Q_W"] = 16314.0
        accessory_data["Q_X"] = -1476.0
        accessory_data["Q_Y"] = 219.0
        accessory_data["Q_Z"] = 217.0
        accessory_data["Acc_X"] = 0.02
        accessory_data["Acc_Y"] = -0.23
        accessory_data["Acc_Z"] = 0.06
        accessory_data["IMU1"] = 0.0
        accessory_data["IMU2"] = 0.0
        accessory_data["Control_flags"] = 130.0
        accessory_data["Control_flags2"] = 0.0
        accessory_data["StatusFlags"] = 6144.0
        accessory_data["Heater"] = 0.0
        accessory_data["MsgCount"] = 69.0
        accessory_data["RdgCount"] = 0.0
        accessory_data["Watchdog"] = 3590.0
        accessory_data["gps_time"] = pd.to_datetime(accessory_data["pc_time_end_measurement"]).dt.tz_localize(None).astype(str)
        accessory_data["cpu_time"] = pd.to_datetime(accessory_data["pc_time_end_measurement"]).dt.tz_localize(None).astype(str)
        accessory_data["dataseries_id"] = self.generate_dataseries_id()
        accessory_data["sample_id"] =  self.generate_sample_ids(accessory_data["pc_time_end_measurement"])
        
        self.accessory_data = accessory_data
    
    
    """
    ---------------------------------------------------------------------------
    individual column functions
    ---------------------------------------------------------------------------
    """
    
    
    def generate_sample_ids(self, timeseries):
        """must have spectral_data generated first"""
        new_sample_ids = pd.DataFrame()
        new_sample_ids["pc_time_end_measurement"] = timeseries
        
        
        if self.spectral_data is not None:
            new_sample_ids = new_sample_ids.merge(self.spectral_data[["pc_time_end_measurement", "sample_id"]], how="left", on="pc_time_end_measurement")
        
        if "sample_id" in new_sample_ids.columns:
            not_matched = pd.isnull(new_sample_ids["sample_id"])
            new_sample_ids.loc[not_matched, "sample_id"] = [str(uuid.uuid1()) for i in range(np.sum(not_matched))]
        else:
            new_sample_ids["sample_id"] = [str(uuid.uuid1()) for i in range(len(timeseries))]
        
        return new_sample_ids["sample_id"]
    
    def generate_dataseries_id(self):
        if self.dataseries_id is None:
            self.dataseries_id = str(uuid.uuid1())
        
        return self.dataseries_id
    
    def linear_integral(self, timeseries):
        return np.ones(len(timeseries))

    def dist_from_midday(self, timeseries):
        timeseries = pd.to_datetime(timeseries)
        mins = timeseries.dt.hour*60 + timeseries.dt.minute
        return 720-np.abs(mins-720)

    def pvlib_integral_static(self, timeseries, *args):
        location = pvlib.location.Location(self.latitude, 
                                           self.longitude, 
                                           altitude=self.altitude)
        integrals = location.get_clearsky(pd.DatetimeIndex(timeseries))
        data = integrals[["ghi", "dhi"]].values*SCALING_FACTOR
        
        return data
    
    def smarts_integral_data(self, timeseries, filepath=None):
        if filepath is None:
            res = importlib.resources.files("hsr1.data").joinpath("smarts.txt")
            file = importlib.resources.as_file(res)
            with file as f:
                filepath = f
        
        smarts_data = spectrumUtils.read_simple_file(filepath)
        
        return smarts_data*len(timeseries)
    
    def integral_hour_of_day(self, timeseries):
        timeseries = pd.to_datetime(timeseries)
        return np.array([timeseries.dt.hour*100, timeseries.dt.hour*50]).T
    
    def remove_n_hour(self, timeseries, hour):
        timestamps = pd.to_datetime(timeseries)
        return timestamps.dt.hour != hour
    
    def num_readings(self, timeseries, base=0, scale=1):
        data = [i for i in range(len(timeseries))]
        data = base + np.array(data)*scale
        return data
        
        
"""
GUIDE TO WRITING CUSTOM FUNCTIONS:

dataseries and timeseries is automatically passed in, so the function must expect it
return a list, np array or series of data for the specific column

use it with

dataseries.generate_custom_column(table, column, function, **kwargs)
"""





if __name__ == "__main__":
    db_name = "databases/synthetic.db"
    
    if os.path.exists(db_name):
        os.remove(db_name)
    
    latitude = 13.19
    longitude = -59.54
    altitude = 40
    
    synthetic_dataset = SyntheticDataset(start_date="2023-01-01 00:00:00+00:00", 
                                         end_date="2023-01-03 23:59:00+00:00",
                                         latitude=latitude, longitude=longitude,
                                         altitude=40)
    
    synthetic_dataset.generate_spectral_data("10s")
    synthetic_dataset.generate_system_data()
    synthetic_dataset.generate_deployment_metadata("C:/Users/albie/work/Datasets/Winster 2023/HSR1-002 Winster 2024 Deployment.ini")
    synthetic_dataset.generate_accessory_data()
    
    synthetic_dataset.generate_custom_column("spectral_data", 
                                             ["global_integral", "diffuse_integral"], 
                                             synthetic_dataset.pvlib_integral_static)
    
    dfs = synthetic_dataset.get_dfs()
    
    db = hsr1.DBDriver(db_name)
    db.store(dfs)
    graph = hsr1.Graph(db)
    graph.plot_integral()
    graph.plot_accessory()
    graph.plot_gps()
    graph.plot_dips_summary()


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License (LGPL)
"""

__version__ = "1.1.0"

from hsr1.dbDriver import DBDriver
from hsr1.graph import Graph

import hsr1.synthetic_dataset
import hsr1.read_txt

